"""Task: coroutine wrapper"""

from __future__ import annotations

import heapq
import logging
from abc import ABC
from collections import Counter, deque
from collections.abc import Awaitable, Callable, Coroutine, Generator
from enum import IntEnum, auto
from typing import Any

from ._loop_if import LoopIf

logger = logging.getLogger("deltacycle")

type Predicate = Callable[[], bool]


class CancelledError(Exception):
    """Task has been cancelled."""


class InvalidStateError(Exception):
    """Task has an invalid state."""


class TaskState(IntEnum):
    """Task State

    Transitions::

        INIT -> RUNNING -> RESULTED
                        -> CANCELLED
                        -> EXCEPTED
    """

    # Initialized
    INIT = auto()

    # Currently running
    RUNNING = auto()

    # Done: returned a result
    RESULTED = auto()
    # Done: cancelled
    CANCELLED = auto()
    # Done: raised an exception
    EXCEPTED = auto()


_task_state_transitions = {
    TaskState.INIT: {TaskState.RUNNING},
    TaskState.RUNNING: {
        TaskState.RESULTED,
        TaskState.CANCELLED,
        TaskState.EXCEPTED,
    },
}


class TaskQueueIf(ABC):
    def __bool__(self) -> bool:
        """Return True if the queue has tasks ready to run."""
        raise NotImplementedError()  # pragma: no cover

    def push(self, item: Any) -> None:
        raise NotImplementedError()  # pragma: no cover

    def pop(self) -> Any:
        raise NotImplementedError()  # pragma: no cover

    def drop(self, task: Task) -> None:
        """If a task reneges, drop it from the queue."""
        raise NotImplementedError()  # pragma: no cover


class PendQueue(TaskQueueIf):
    """Priority queue for ordering task execution."""

    def __init__(self):
        # time, priority, index, task, value
        self._items: list[tuple[int, int, int, Task, Any]] = []

        # Monotonically increasing integer
        # Breaks (time, priority, ...) ties in the heapq
        self._index: int = 0

    def __bool__(self) -> bool:
        return bool(self._items)

    def push(self, item: tuple[int, Task, Any]):
        time, task, value = item
        task._link(self)
        heapq.heappush(self._items, (time, task.priority, self._index, task, value))
        self._index += 1

    def pop(self) -> tuple[Task, Any]:
        _, _, _, task, value = heapq.heappop(self._items)
        task._unlink(self)
        return (task, value)

    def drop(self, task: Task):
        for i, (_, _, _, t, _) in enumerate(self._items):
            if t is task:
                index = i
                break
        else:
            assert False  # pragma: no cover
        self._items.pop(index)

    def peek(self) -> int:
        return self._items[0][0]

    def clear(self):
        while self._items:
            self.pop()
        self._index = 0


class WaitFifo(TaskQueueIf):
    """Tasks wait in FIFO order."""

    def __init__(self):
        self._items: deque[Task] = deque()

    def __bool__(self) -> bool:
        return bool(self._items)

    def push(self, item: Task):
        task = item
        task._link(self)
        self._items.append(task)

    def pop(self) -> Task:
        task = self._items.popleft()
        task._unlink(self)
        return task

    def drop(self, task: Task):
        self._items.remove(task)


class WaitTouch(TaskQueueIf):
    """Tasks wait for variable touch."""

    def __init__(self):
        self._items: deque[Task] = deque()
        self._tps: dict[Task, Predicate] = dict()

    def __bool__(self) -> bool:
        return bool(self._items)

    def push(self, item: tuple[Predicate, Task]):
        p, task = item
        task._link(self)
        self._tps[task] = p

    def pop(self) -> Task:
        task = self._items.popleft()
        task._unlink(self)
        return task

    def drop(self, task: Task):
        del self._tps[task]

    def touch(self):
        assert not self._items
        for task, p in self._tps.items():
            if p():
                self._items.append(task)


class Task(Awaitable[Any], LoopIf):
    """Coroutine wrapper."""

    _index = 0

    def __init__(
        self,
        coro: Coroutine[Any, Any, Any],
        name: str | None = None,
        priority: int = 0,
    ):
        self._state = TaskState.INIT

        self._coro = coro
        if name is None:
            self._name = f"Task-{self.__class__._index}"
            self.__class__._index += 1
        else:
            self._name = name
        self._priority = priority

        # Reference counts for this task
        self._refcnts: Counter[TaskQueueIf] = Counter()

        # Other tasks waiting for this task to complete
        self._waiting = WaitFifo()

        # Completion
        self._result: Any = None

        # Exception
        self._exception: Exception | None = None

    def __await__(self) -> Generator[None, Any, Task]:
        if not self.done():
            task = self._loop.task()
            self._waiting.push(task)
            t: Task = yield from self._loop.switch_gen()
            assert t is self

        # Resume
        return self.result()

    def _wait(self, task: Task):
        self._waiting.push(task)

    def _set(self):
        while self._waiting:
            task = self._waiting.pop()
            self._loop.call_soon(task, value=self)

    @property
    def coro(self) -> Coroutine[Any, Any, Any]:
        return self._coro

    @property
    def name(self) -> str:
        return self._name

    @property
    def priority(self) -> int:
        return self._priority

    def _set_state(self, state: TaskState):
        assert state in _task_state_transitions[self._state]
        logger.debug("%s: %s => %s", self.name, self._state.name, state.name)
        self._state = state

    def state(self) -> TaskState:
        return self._state

    def _link(self, tq: TaskQueueIf):
        self._refcnts[tq] += 1

    def _unlink(self, tq: TaskQueueIf):
        assert self._refcnts[tq] > 0
        self._refcnts[tq] -= 1

    def _renege(self):
        while self._refcnts:
            tq, cnt = self._refcnts.popitem()
            for _ in range(cnt):
                tq.drop(self)

    def _do_run(self, value: Any = None):
        if self._state is TaskState.INIT:
            self._set_state(TaskState.RUNNING)
        elif self._state is not TaskState.RUNNING:
            assert False  # pragma: no cover

        # Start / Resume coroutine
        if self._exception is None:
            self._coro.send(value)
        else:
            self._coro.throw(self._exception)

    def _do_result(self, exc: StopIteration):
        self._set_result(exc.value)
        self._set_state(TaskState.RESULTED)
        self._set()

    def _do_cancel(self, exc: CancelledError):
        self._set_exception(exc)
        self._set_state(TaskState.CANCELLED)
        self._set()

    def _do_except(self, exc: Exception):
        self._set_exception(exc)
        self._set_state(TaskState.EXCEPTED)
        self._set()

    _done_states = frozenset([TaskState.RESULTED, TaskState.CANCELLED, TaskState.EXCEPTED])

    def done(self) -> bool:
        """Return True if the task is done.

        A task that is "done" either 1) completed normally,
        2) was cancelled by another task, or 3) raised an exception.
        """
        return self._state in self._done_states

    def _set_result(self, result: Any):
        if self.done():
            raise InvalidStateError("Task is already done")
        self._result = result

    def result(self) -> Any:
        """Return the task's result, or raise an exception.

        Returns:
            If the task ran to completion, return its result.

        Raises:
            CancelledError: If the task was cancelled.
            Exception: If the task raise any other type of exception.
            InvalidStateError: If the task is not done.
        """
        if self._state is TaskState.RESULTED:
            assert self._exception is None
            return self._result
        if self._state is TaskState.CANCELLED:
            assert isinstance(self._exception, CancelledError)
            raise self._exception
        if self._state is TaskState.EXCEPTED:
            assert isinstance(self._exception, Exception)
            raise self._exception
        raise InvalidStateError("Task is not done")

    def _set_exception(self, exc: Exception):
        if self.done():
            raise InvalidStateError("Task is already done")
        self._exception = exc

    def exception(self) -> Exception | None:
        """Return the task's exception.

        Returns:
            If the task raised an exception, return it.
            Otherwise, return None.

        Raises:
            If the task was cancelled, re-raise the CancelledError.
        """
        if self._state is TaskState.RESULTED:
            assert self._exception is None
            return self._exception
        if self._state is TaskState.CANCELLED:
            assert isinstance(self._exception, CancelledError)
            raise self._exception
        if self._state is TaskState.EXCEPTED:
            assert isinstance(self._exception, Exception)
            return self._exception
        raise InvalidStateError("Task is not done")

    def cancel(self, msg: str | None = None) -> bool:
        """Schedule task for cancellation.

        If a task is already done: return False.

        If a task is pending or waiting:

        1. Renege from all queues
        2. Reschedule to raise CancelledError in the current time slot
        3. Return True

        If a task is running, immediately raise CancelledError.

        Args:
            msg: Optional str message passed to CancelledError instance

        Returns:
            bool success indicator

        Raises:
            CancelledError: If the task cancels itself
        """
        # Already done; do nothing
        if self.done():
            return False

        args = () if msg is None else (msg,)
        exc = CancelledError(*args)

        # Task is cancelling itself. Weird, but legal.
        if self is self._loop.task():
            raise exc

        # Pending/Waiting tasks must first renege from queues
        self._renege()

        # Reschedule for cancellation
        self._set_exception(exc)
        self._loop.call_soon(self)

        return True



"""Test basic loop functionality"""

import logging

import pytest
from pytest import LogCaptureFixture

from deltacycle import get_loop, get_running_loop, irun, run, set_loop, sleep

logger = logging.getLogger("deltacycle")


async def main(n: int):
    for i in range(n):
        logger.info("%d", i)
        await sleep(1)
    return n


def test_run(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    ret = run(main(42))
    assert ret == 42

    msgs = [(r.time, r.getMessage()) for r in caplog.records]
    assert msgs == [(i, str(i)) for i in range(42)]


def test_irun(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    g = irun(main(42))
    try:
        while True:
            next(g)
    except StopIteration as e:
        assert e.value == 42

    msgs = [(r.time, r.getMessage()) for r in caplog.records]
    assert msgs == [(i, str(i)) for i in range(42)]


def test_cannot_run(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    run(main(100))
    loop = get_loop()

    # Loop is already in COMPLETED state
    with pytest.raises(RuntimeError):
        run(loop=loop)

    with pytest.raises(RuntimeError):
        list(irun(loop=loop))


def test_limits(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    run(main(1000), ticks=51)
    loop = get_running_loop()
    assert loop.time() == 50

    run(loop=loop, ticks=51)
    assert loop.time() == 100

    run(loop=loop, until=201)
    assert loop.time() == 200

    with pytest.raises(TypeError):
        run(loop=loop, ticks=42, until=42)


def test_nocoro():
    with pytest.raises(ValueError):
        run()
    with pytest.raises(ValueError):
        list(irun())


def test_get_running_loop():
    # No loop
    set_loop()
    with pytest.raises(RuntimeError):
        get_running_loop()

    # Loop is not running
    run(main(42))
    with pytest.raises(RuntimeError):
        get_running_loop()



"""Task Group"""

from collections.abc import Coroutine
from types import TracebackType
from typing import Any

from ._loop_if import LoopIf
from ._task import Task, TaskState


class TaskGroup(LoopIf):
    """Group of tasks."""

    def __init__(self):
        self._parent = self._loop.task()
        # Preserve child task order
        self._children: list[Task] = []
        self._done: list[Task] = []
        self._excs: list[Exception] = []

    async def __aenter__(self):
        return self

    def _record(self, child: Task) -> bool:
        self._done.append(child)
        if child.state() in {TaskState.RESULTED, TaskState.CANCELLED}:
            return False
        if child.state() is TaskState.EXCEPTED:
            self._excs.append(child._exception)  # type: ignore
            return True
        assert False  # pragma: no cover

    async def __aexit__(
        self,
        exc_type: type[Exception] | None,
        exc: Exception | None,
        traceback: TracebackType | None,
    ):
        not_done: set[Task] = set()

        for child in self._children:
            if child.done():
                # NOTE: Ignore child exception
                _ = self._record(child)
            else:
                not_done.add(child)
                child._wait(self._parent)

        # Parent raised an exception
        if exc:
            # Cancel all children
            for child in not_done:
                child.cancel()
            while not_done:
                child: Task = await self._loop.switch_coro()
                not_done.remove(child)
                # NOTE: Ignore child exception
                _ = self._record(child)

            assert set(self._children) == set(self._done)

            # Suppress child exception(s)
            # Re-raise parent exception
            return False

        cancelled: set[Task] = set()

        # Parent did NOT raise an exception
        # Wait for children to complete
        while not_done:
            child: Task = await self._loop.switch_coro()
            not_done.remove(child)
            x = self._record(child)

            # If children spawn newborns, add them to the waiting set
            n = len(self._done) + len(not_done)
            for newborn in self._children[n:]:
                not_done.add(newborn)
                newborn._wait(self._parent)

            # If child raised an exception, cancel remaining siblings
            if x:
                for sibling in not_done - cancelled:
                    sibling.cancel()
                    cancelled.add(sibling)

        assert set(self._children) == set(self._done)

        # Re-raise child exception(s)
        if self._excs:
            raise ExceptionGroup("errors", self._excs)

    def create_task(
        self,
        coro: Coroutine[Any, Any, Any],
        name: str | None = None,
        priority: int = 0,
    ) -> Task:
        task = self._loop.create_task(coro, name, priority)
        self._children.append(task)
        return task



"""Queue synchronization primitive."""

from collections import deque
from collections.abc import Sized
from typing import Any

from ._loop_if import LoopIf
from ._task import WaitFifo


class Queue(Sized, LoopIf):
    """First-in, First-out (FIFO) queue."""

    def __init__(self, maxlen: int = 0):
        self._maxlen = maxlen
        self._items: deque[Any] = deque()
        self._wait_not_empty = WaitFifo()
        self._wait_not_full = WaitFifo()

    def __len__(self) -> int:
        return len(self._items)

    def empty(self) -> bool:
        return not self._items

    def full(self) -> bool:
        return self._maxlen > 0 and len(self._items) == self._maxlen

    def _put(self, item: Any):
        self._items.append(item)
        if self._wait_not_empty:
            self._loop.call_soon(self._wait_not_empty.pop(), value=self)

    def try_put(self, item: Any) -> bool:
        if self.full():
            return False
        self._put(item)
        return True

    async def put(self, item: Any):
        if self.full():
            task = self._loop.task()
            self._wait_not_full.push(task)
            await self._loop.switch_coro()

        self._put(item)

    def _get(self) -> Any:
        item = self._items.popleft()
        if self._wait_not_full:
            self._loop.call_soon(self._wait_not_full.pop(), value=self)
        return item

    def try_get(self) -> tuple[bool, Any]:
        if self.empty():
            return False, None
        return True, self._get()

    async def get(self) -> Any:
        if self.empty():
            task = self._loop.task()
            self._wait_not_empty.push(task)
            await self._loop.switch_coro()

        return self._get()



"""Event Loop"""

from __future__ import annotations

import logging
from collections.abc import Awaitable, Callable, Coroutine, Generator
from enum import IntEnum, auto
from typing import Any

from ._task import CancelledError, PendQueue, Task
from ._variable import Variable

logger = logging.getLogger("deltacycle")

type Predicate = Callable[[], bool]


class _SuspendResume(Awaitable[Any]):
    """Suspend/Resume current task.

    Use case:
    1. Current task A suspends itself: RUNNING => WAITING
    2. Event loop chooses PENDING tasks ..., T
    3. ... Task T wakes up task A w/ value X: WAITING => PENDING
    4. Event loop chooses PENDING tasks ..., A: PENDING => RUNNING
    5. Task A resumes with value X

    The value X can be used to pass information to the task.
    """

    def __await__(self) -> Generator[None, Any, Any]:
        # Suspend
        value = yield
        # Resume
        return value


class _FinishError(Exception):
    """Force the simulation to stop."""


def create_task(
    coro: Coroutine[Any, Any, Any],
    name: str | None = None,
    priority: int = 0,
) -> Task:
    """Create a task, and schedule it to start soon."""
    loop = get_running_loop()
    return loop.create_task(coro, name, priority)


class LoopState(IntEnum):
    """Loop State

    Transitions::

        INIT -> RUNNING -> COMPLETED
                        -> FINISHED
    """

    # Initialized
    INIT = auto()

    # Currently running
    RUNNING = auto()

    # All tasks completed
    COMPLETED = auto()

    # finish() called
    FINISHED = auto()


_loop_state_transitions = {
    LoopState.INIT: {LoopState.RUNNING},
    LoopState.RUNNING: {LoopState.COMPLETED, LoopState.FINISHED},
}


class Loop:
    """Simulation event loop.

    Responsible for:

    * Scheduling and executing all tasks
    * Updating all model state

    This is a low level API.
    User code is not expected to interact with it directly.
    To run a simulation, use the run and irun functions.
    """

    _index = 0

    init_time = -1
    start_time = 0

    main_priority = 0

    def __init__(self):
        self._name = f"Loop-{self.__class__._index}"
        self.__class__._index += 1

        self._state = LoopState.INIT

        # Simulation time
        self._time: int = self.init_time

        # Main task
        self._main: Task | None = None

        # Currently executing task
        self._task: Task | None = None

        # Task queue
        self._queue = PendQueue()

        # Model variables
        self._touched: set[Variable] = set()

    def _set_state(self, state: LoopState):
        assert state in _loop_state_transitions[self._state]
        logger.debug("%s: %s => %s", self._name, self._state.name, state.name)
        self._state = state

    def state(self) -> LoopState:
        """Current simulation state."""
        return self._state

    def time(self) -> int:
        """Current simulation time."""
        return self._time

    @property
    def main(self) -> Task:
        """Parent task of all other tasks."""
        assert self._main is not None
        return self._main

    def task(self) -> Task:
        """Currently running task."""
        assert self._task is not None
        return self._task

    _done_states = frozenset([LoopState.COMPLETED, LoopState.FINISHED])

    def done(self) -> bool:
        return self._state in self._done_states

    def finished(self) -> bool:
        return self._state is LoopState.FINISHED

    # Scheduling methods
    def call_soon(self, task: Task, value: Any = None):
        self._queue.push((self._time, task, value))

    def call_later(self, delay: int, task: Task, value: Any = None):
        self._queue.push((self._time + delay, task, value))

    def call_at(self, when: int, task: Task, value: Any = None):
        self._queue.push((when, task, value))

    def create_main(self, coro: Coroutine[Any, Any, Any]) -> Task:
        assert self._time == self.init_time
        main = Task(coro, name="main", priority=self.main_priority)
        self._main = main
        self.call_at(self.start_time, main)
        return main

    def create_task(
        self,
        coro: Coroutine[Any, Any, Any],
        name: str | None = None,
        priority: int = 0,
    ) -> Task:
        assert self._time >= self.start_time
        task = Task(coro, name=name, priority=priority)
        self.call_soon(task)
        return task

    async def switch_coro(self) -> Any:
        assert self._task is not None
        # Suspend
        value = await _SuspendResume()
        # Resume
        return value

    def switch_gen(self) -> Generator[None, Any, Any]:
        assert self._task is not None
        # Suspend
        value = yield
        # Resume
        return value

    def touch(self, v: Variable):
        self._touched.add(v)

    def _update(self):
        while self._touched:
            v = self._touched.pop()
            v.update()

    def _finish(self):
        self._queue.clear()
        self._touched.clear()
        self._set_state(LoopState.FINISHED)

    def _iter_time_slot(self, time: int) -> Generator[tuple[Task, Any], None, None]:
        """Iterate through all tasks in a time slot.

        The first task has already been peeked.
        This is a do-while loop.
        """
        task, value = self._queue.pop()
        yield (task, value)
        while self._queue and self._queue.peek() == time:
            task, value = self._queue.pop()
            yield (task, value)

    def _kernel(self, limit: int | None):
        if self._state is LoopState.INIT:
            self._set_state(LoopState.RUNNING)
        elif self._state is not LoopState.RUNNING:
            s = f"Loop has invalid state: {self._state.name}"
            raise RuntimeError(s)

        while self._queue:
            # Peek when next event is scheduled
            time = self._queue.peek()

            # Protect against time traveling tasks
            assert time > self._time

            # Halt if we hit the run limit
            if limit is not None and time >= limit:
                return

            # Otherwise, advance to new timeslot
            self._time = time

            # Execute time slot
            for task, value in self._iter_time_slot(time):
                self._task = task
                try:
                    task._do_run(value)
                except StopIteration as e:
                    task._do_result(e)
                except CancelledError as e:
                    task._do_cancel(e)
                except _FinishError:
                    self._finish()
                    return
                except Exception as e:
                    task._do_except(e)

            # Update simulation state
            self._update()

        # All tasks exhausted
        self._set_state(LoopState.COMPLETED)

    def run(self, ticks: int | None = None, until: int | None = None):
        # Determine the run limit
        match ticks, until:
            # Run until no tasks left
            case None, None:
                limit = None
            # Run until an absolute time
            case None, int():
                limit = until
            # Run until a number of ticks in the future
            case int(), None:
                limit = max(self.start_time, self._time) + ticks
            case _:
                s = "Expected either ticks or until to be int | None"
                raise TypeError(s)

        self._kernel(limit)

    def __iter__(self) -> Generator[int, None, None]:
        if self._state is LoopState.INIT:
            self._set_state(LoopState.RUNNING)
        elif self._state is not LoopState.RUNNING:
            s = f"Loop has invalid state: {self._state.name}"
            raise RuntimeError(s)

        while self._queue:
            # Peek when next event is scheduled
            time = self._queue.peek()

            # Protect against time traveling tasks
            assert time > self._time

            # Yield before entering new timeslot
            yield time

            # Advance to new timeslot
            self._time = time

            # Execute time slot
            for task, value in self._iter_time_slot(time):
                self._task = task
                try:
                    task._do_run(value)
                except StopIteration as e:
                    task._do_result(e)
                except CancelledError as e:
                    task._do_cancel(e)
                except _FinishError:
                    self._finish()
                    return
                except Exception as e:
                    task._do_except(e)

            # Update simulation state
            self._update()

        # All tasks exhausted
        self._set_state(LoopState.COMPLETED)


_loop: Loop | None = None


def get_running_loop() -> Loop:
    """Return currently running loop.

    Returns:
        Loop instance

    Raises:
        RuntimeError: No loop, or loop is not currently running.
    """
    if _loop is None:
        raise RuntimeError("No loop")
    if _loop.state() is not LoopState.RUNNING:
        raise RuntimeError("Loop not RUNNING")
    return _loop


def get_loop() -> Loop | None:
    """Get the current event loop."""
    return _loop


def set_loop(loop: Loop | None = None):
    """Set the current event loop."""
    global _loop  # noqa: PLW0603
    _loop = loop


def get_current_task() -> Task:
    """Return currently running task.

    Returns:
        Task instance

    Raises:
        RuntimeError: No loop, or loop is not currently running.
    """
    loop = get_running_loop()
    return loop.task()


def now() -> int:
    """Return current simulation time."""
    loop = get_running_loop()
    return loop.time()


def _run_pre(coro: Coroutine[Any, Any, Any] | None, loop: Loop | None) -> Loop:
    if loop is None:
        set_loop(loop := Loop())
        if coro is None:
            raise ValueError("New loop requires a valid coro arg")
        assert coro is not None
        _ = loop.create_main(coro)
    else:
        set_loop(loop)
    return loop


def run(
    coro: Coroutine[Any, Any, Any] | None = None,
    loop: Loop | None = None,
    ticks: int | None = None,
    until: int | None = None,
) -> Any:
    """Run a simulation.

    If a simulation hits the run limit, it will exit and return None.
    That simulation may be resumed any number of times.
    If all tasks are exhausted, return the main coroutine result.

    Args:
        coro: Optional main coroutine.
            Required if creating a new loop.
            Ignored if using an existing loop.
        loop: Optional Loop instance.
            If not provided, a new loop will be created.
        ticks: Optional relative run limit.
            If provided, run for *ticks* simulation time steps.
        until: Optional absolute run limit.
            If provided, run until *ticks* simulation time steps.

    Returns:
        If the main coroutine runs til completion, return its result.
        Otherwise, return ``None``.

    Raises:
        ValueError: Creating a new loop, but no coro provided.
        TypeError: ticks and until args conflict.
        RuntimeError: The loop is in an invalid state.
    """
    loop = _run_pre(coro, loop)
    loop.run(ticks, until)

    if loop.main.done():
        return loop.main.result()


def irun(
    coro: Coroutine[Any, Any, Any] | None = None,
    loop: Loop | None = None,
) -> Generator[int, None, Any]:
    """Iterate a simulation.

    Iterated simulations do not have a run limit.
    It is the user's responsibility to break at the desired time.
    If all tasks are exhausted, return the main coroutine result.

    Args:
        coro: Optional main coroutine.
            Required if creating a new loop.
            Ignored if using an existing loop.
        loop: Optional Loop instance.
            If not provided, a new loop will be created.

    Yields:
        int time immediately *before* the next time slot executes.

    Returns:
        main coroutine result.

    Raises:
        ValueError: Creating a new loop, but no coro provided.
        TypeError: ticks and until args conflict.
        RuntimeError: The loop is in an invalid state.
    """
    loop = _run_pre(coro, loop)
    yield from loop

    assert loop.main.done()
    return loop.main.result()


async def sleep(delay: int):
    """Suspend the task, and wake up after a delay."""
    loop = get_running_loop()
    task = loop.task()
    loop.call_later(delay, task)
    await _SuspendResume()


async def changed(*vs: Variable) -> Variable:
    """Resume execution upon variable change.

    Suspend execution of the current task;
    Resume when any variable in the sensitivity list changes.

    Args:
        vs: Tuple of Variables, a sensitivity list.

    Returns:
        The Variable instance that triggered the task to resume.
    """
    loop = get_running_loop()
    task = loop.task()
    for v in vs:
        v._wait(v.changed, task)
    v: Variable = await loop.switch_coro()
    return v


async def touched(vps: dict[Variable, Predicate | None]) -> Variable:
    """Resume execution upon predicated variable change.

    Suspend execution of the current task;
    Resume when any variable in the sensitivity list changes,
    *and* the predicate function evaluates to True.
    If the predicate function is None, it will default to *any* change.

    Args:
        vps: Dict of Variable => Predicate mappings, a sensitivity list.

    Returns:
        The Variable instance that triggered the task to resume.
    """
    loop = get_running_loop()
    task = loop.task()
    for v, p in vps.items():
        if p is None:
            v._wait(v.changed, task)
        else:
            v._wait(p, task)
    v: Variable = await loop.switch_coro()
    return v


def finish():
    """Halt all incomplete coroutines, and immediately exit simulation.

    Clear all loop data, and transition state to FINISHED.
    """
    raise _FinishError()



"""Event synchronization primitive"""

from ._loop_if import LoopIf
from ._task import WaitFifo


class Event(LoopIf):
    """Notify multiple tasks that some event has happened."""

    def __init__(self):
        self._flag = False
        self._waiting = WaitFifo()

    async def wait(self):
        if not self._flag:
            task = self._loop.task()
            self._waiting.push(task)
            await self._loop.switch_coro()

    def set(self):
        while self._waiting:
            self._loop.call_soon(self._waiting.pop(), value=self)
        self._flag = True

    def clear(self):
        self._flag = False

    def is_set(self) -> bool:
        return self._flag



"""Test deltacycle.Event"""

import logging

from pytest import LogCaptureFixture

from deltacycle import Event, create_task, run, sleep

logger = logging.getLogger("deltacycle")


async def primary(event: Event):
    logger.info("enter")

    await sleep(10)

    # T=10
    logger.info("set")
    event.set()
    assert event.is_set()

    await sleep(10)

    # T=20
    logger.info("clear")
    event.clear()
    assert not event.is_set()

    await sleep(10)

    # T=30
    logger.info("set")
    event.set()
    assert event.is_set()

    logger.info("exit")


async def secondary(event: Event):
    logger.info("enter")

    # Event clear
    logger.info("waiting")
    await event.wait()

    # Event set @10
    logger.info("running")
    await sleep(10)

    # Event clear
    logger.info("waiting")
    await event.wait()

    # Event set @30
    logger.info("running")
    await sleep(10)

    # Event still set: return immediately
    await event.wait()

    logger.info("exit")


EXP1 = {
    # P
    (0, "P", "enter"),
    (10, "P", "set"),
    (20, "P", "clear"),
    (30, "P", "set"),
    (30, "P", "exit"),
    # S1
    (0, "S1", "enter"),
    (0, "S1", "waiting"),
    (10, "S1", "running"),
    (20, "S1", "waiting"),
    (30, "S1", "running"),
    (40, "S1", "exit"),
    # S2
    (0, "S2", "enter"),
    (0, "S2", "waiting"),
    (10, "S2", "running"),
    (20, "S2", "waiting"),
    (30, "S2", "running"),
    (40, "S2", "exit"),
    # S3
    (0, "S3", "enter"),
    (0, "S3", "waiting"),
    (10, "S3", "running"),
    (20, "S3", "waiting"),
    (30, "S3", "running"),
    (40, "S3", "exit"),
}


def test_acquire_release(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        event = Event()
        create_task(primary(event), name="P")
        create_task(secondary(event), name="S1")
        create_task(secondary(event), name="S2")
        create_task(secondary(event), name="S3")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1



"""Model variables"""

from __future__ import annotations

from abc import ABC
from collections import defaultdict
from collections.abc import Awaitable, Generator, Hashable
from typing import Any

from ._loop_if import LoopIf
from ._task import Predicate, Task, WaitTouch


class Variable(Awaitable[Any], LoopIf):
    """Model component.

    Children::

               Variable
                  |
           +------+------+
           |             |
        Singular     Aggregate
    """

    def __init__(self):
        self._waiting = WaitTouch()

    def __await__(self) -> Generator[None, Any, Variable]:
        task = self._loop.task()
        self._waiting.push((self.changed, task))
        v: Variable = yield from self._loop.switch_gen()
        assert v is self
        return v

    def _wait(self, p: Predicate, task: Task):
        self._waiting.push((p, task))

    def _set(self):
        self._waiting.touch()

        while self._waiting:
            task = self._waiting.pop()
            task._renege()
            self._loop.call_soon(task, value=self)

        # Add variable to update set
        self._loop.touch(self)

    value = property(fget=lambda self: NotImplemented)

    def changed(self) -> bool:
        """Return True if changed during the current time slot."""
        raise NotImplementedError()  # pragma: no cover

    def update(self) -> None:
        """Loop callback."""
        raise NotImplementedError()  # pragma: no cover


class Value(ABC):
    """Variable value."""

    def _get_prev(self) -> Any:
        raise NotImplementedError()  # pragma: no cover

    prev = property(fget=_get_prev)

    def _set_next(self, value: Any) -> None:
        raise NotImplementedError()  # pragma: no cover

    next = property(fset=_set_next)


class Singular(Variable, Value):
    """Model state organized as a single unit."""

    def __init__(self, value: Any):
        Variable.__init__(self)
        self._prev = value
        self._next = value
        self._changed: bool = False

    # Value
    def _get_prev(self) -> Any:
        return self._prev

    prev = property(fget=_get_prev)

    def _set_next(self, value: Any):
        self._changed = value != self._next
        self._next = value

        # Notify the event loop
        self._set()

    next = property(fset=_set_next)

    # Variable
    def _get_value(self):
        return self._next

    value = property(fget=_get_value)

    def changed(self) -> bool:
        return self._changed

    def update(self):
        self._prev = self._next
        self._changed = False


class Aggregate(Variable):
    """Model state organized as multiple units."""

    def __init__(self, value: Any):
        Variable.__init__(self)
        self._prevs: dict[Hashable, Any] = defaultdict(lambda: value)
        self._nexts: dict[Hashable, Any] = dict()

    # [key] => Value
    def __getitem__(self, key: Hashable) -> AggrItem:
        return AggrItem(self, key)

    def _get_prev(self, key: Hashable) -> Any:
        return self._prevs[key]

    def _get_next(self, key: Hashable) -> Any:
        try:
            return self._nexts[key]
        except KeyError:
            return self._prevs[key]

    def _set_next(self, key: Hashable, value: Any):
        if value != self._get_next(key):
            self._nexts[key] = value

        # Notify the event loop
        self._set()

    # Variable
    def _get_value(self) -> AggrValue:
        return AggrValue(self)

    value = property(fget=_get_value)

    def changed(self) -> bool:
        return bool(self._nexts)

    def update(self):
        while self._nexts:
            key, value = self._nexts.popitem()
            self._prevs[key] = value


class AggrItem(Value):
    """Wrap Aggregate __getitem__."""

    def __init__(self, aggr: Aggregate, key: Hashable):
        self._aggr = aggr
        self._key = key

    def _get_prev(self) -> Any:
        return self._aggr._get_prev(self._key)

    prev = property(fget=_get_prev)

    def _set_next(self, value: Any):
        self._aggr._set_next(self._key, value)

    next = property(fset=_set_next)


class AggrValue:
    """Wrap Aggregate value."""

    def __init__(self, aggr: Aggregate):
        self._aggr = aggr

    def __getitem__(self, key: Hashable) -> Any:
        return self._aggr._get_next(key)



"""LoopIf abstract base class

Allows easy access to global event loop for Event, Semaphore, Task, ...
Works around tricky circular import: Loop => Task => Loop.
"""

from abc import ABC
from functools import cached_property


class LoopIf(ABC):
    @cached_property
    def _loop(self):
        from ._loop import get_running_loop

        return get_running_loop()
